//------------------------------------------------------------------------------
// Desc   : Generic fake BRDF shader 
// Author : 
//    Maya CgFX: Cedric Caillaud (http://www.kjapi.com) & 
//               Brice Vandemoortele (http://www.mentalwarp.com)
//    3dsmax FX: ported by Christoph Kubisch (http://www.luxinia.de)
// Date   : October 2007
// Requirements:
//    Shader Model 3 capable card (Nvidia GeForce 6, Radeon X1300)
//    3dsmax: only mapping channel 1 will be used for all textures


//------------------------------------------------------------------------------
//This work is licensed under a  Creative Commons Attribution-Share Alike 3.0 Unported License
// http://creativecommons.org/licenses/by-sa/3.0/
//------------------------------------------------------------------------------


//-----------------------------------------------
// Matrices
//-----------------------------------------------
float4x4 wvpMatrix	: WorldViewProjection;
float4x4 wMatrix	: World;
float4x4 wvMatrixI	: WorldViewInverse;

//-----------------------------------------------
// Light position
//-----------------------------------------------
float4 LightPos : Position 
<
	string UIName = "Light";
    string Object = "OmniLight";
    string Space = "World";
    int RefID = 0;
> = {-54.0f, 50.0f, 100.0f, 1.0f};

//-----------------------------------------------
// Light color
//-----------------------------------------------
float4 LightColor : LightColor <
	int LightRef = 0;
> = float4( 1.0f, 1.0f, 1.0f, 1.0f );    // diffuse


//-----------------------------------------------
// BRDF Sampler
//-----------------------------------------------
texture BRDF
<
	string UIName = "Texture BRDF";
>;
 
sampler2D BRDFSampler = sampler_state
{
	Texture = <BRDF>;
	MipFilter = None;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	AddressU = Clamp;
	AddressV = Clamp;
	AddressW = Clamp;
};

half OffsetU
<
	string UIType = "FloatSpinner";
	float UIMin = -1.0;
	float UIMax = 1.0;
	float UIStep = 0.01;
	string UIName = "U Offset";
> = 0.0;

half OffsetV
<
	string UIType = "FloatSpinner";
	float UIMin = -1.0;
	float UIMax = 1.0;
	float UIStep = 0.01;
	string UIName = "V Offset";
> = 0.0;

//-----------------------------------------------
// diffuse sampler - D
//-----------------------------------------------
texture diffuseTexture	: DiffuseMap
< 
	string UIName = "DiffuseMap";
>;
sampler2D diffuseSampler = sampler_state 
{	
	Texture = <diffuseTexture>;
	MinFilter	=	Linear;
	MagFilter	=	Linear;
	MipFilter	=	Linear;
	AddressU	=	WRAP;
	AddressV	=	WRAP;
};

// Textures Enforce channel 1
int texcoord0 : Texcoord
<
	int Texcoord = 0;
	int MapChannel = 1;
>;

// Vertexcolor
int texcoord1 : Texcoord
<
	int Texcoord = 1;
	int MapChannel = 0;
>;

//-----------------------------------------------
// Alpha Translu sampler - A
//-----------------------------------------------
texture AlphaTexture	: AlphaMap
< 
	string UIName = "AlphaMap";
>;
sampler2D AlphaSampler = sampler_state 
{	
	Texture = <AlphaTexture>;
	MinFilter	=	Linear;
	MagFilter	=	Linear;
	MipFilter	=	Linear;
	AddressU	=	WRAP;
	AddressV	=	WRAP;
};

//-----------------------------------------------
// Fake SSS color
//-----------------------------------------------
half3 SSSColor
<
	string type		= "color";
	string UIName	= "SSSColor";
	string UIWidget = "Color";
> = {0, 0, 0};

//-----------------------------------------------
// Normal sampler - N
//-----------------------------------------------
texture NormalTexture : NormalMap
< 
	string UIName = "NormalMap";
>;
sampler2D NormalSampler = sampler_state
{
	Texture = <NormalTexture>;
	MinFilter	=	Linear;
	MagFilter	=	Linear;
	MipFilter	=	Linear;
	AddressU	=	WRAP;
	AddressV	=	WRAP;
};

//-----------------------------------------------
// Bump Factor
//-----------------------------------------------
half BumpFactor
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 3;
	float UIStep	= 0.1;
	string UIName	=  "Bump Scale";
> = 1.0;

//-----------------------------------------------
// specular sampler - S
//-----------------------------------------------
texture SpecularTexture	: SpecularMap
<
	string UIName = "SpecularMap";
>;
sampler2D SpecularSampler = sampler_state
{
	Texture = <SpecularTexture>;
	MinFilter	=	Linear;
	MagFilter	=	Linear;
	MipFilter	=	Linear;
	AddressU	=	WRAP;
	AddressV	=	WRAP;
};

//-----------------------------------------------
// Specular factor
//-----------------------------------------------
half SpecFactor : SpecularFactor 
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 1.0;
	float UIStep	= 0.1;
	string UIName	=  "Specular Factor";
> = 1.0;

//-----------------------------------------------
// Specular Fresnel Power
//-----------------------------------------------
half SpecularFresnelPower
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.01;
	float UIMax		= 10;
	float UIStep	= 0.1;
	string UIName	=  "Specular Fresnel Factor";
> = 0;

//-----------------------------------------------
// specular power sampler - SP
//-----------------------------------------------
texture SpecularPowerTexture	: SpecularPowerMap
<
	string UIName = "SpecularPowerMap";
>;
sampler2D SpecularPowerSampler = sampler_state
{
	Texture = <SpecularPowerTexture>;
	MinFilter	=	Linear;
	MagFilter	=	Linear;
	MipFilter	=	Linear;
	AddressU	=	WRAP;
	AddressV	=	WRAP;
};

//-----------------------------------------------
// Gloss factor
//-----------------------------------------------
half GlossFactor : GlossinessFactor
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 1.0;
	float UIStep	= 0.1;
	string UIName	= "Gloss Factor";
> = 1.0;

//-----------------------------------------------
// emissive sampler - E
//-----------------------------------------------
texture EmissiveTexture	: EmissiveMap
<
	string UIName = "EmissiveMap";
>;
sampler2D EmissiveSampler = sampler_state
{
	Texture = <EmissiveTexture>;
	MinFilter	=	Linear;
	MagFilter	=	Linear;
	MipFilter	=	Linear;
	AddressU	=	WRAP;
	AddressV	=	WRAP;
};

//-----------------------------------------------
//emissive factor
//-----------------------------------------------
half EmissiveFactor
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 10;
	float UIStep	= 0.1;
	string UIName	=  "Emissive Factor";
> = 0;

//-----------------------------------------------
//emissive color
//-----------------------------------------------
half3 EmissiveColor
<
	string type		= "color";
	string UIName	= "EmissiveColor";
	string UIWidget = "Color";
> = {1.0, 1.0, 1.0};

//-----------------------------------------------
// Env sampler
//-----------------------------------------------
texture EnvTexture
<
	string UIName = "EnvTexture";
	string type = "CUBE";
>;
samplerCUBE EnvSampler = sampler_state
{
	Texture = <EnvTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	//AddressU = Clamp;
	//AddressV = Clamp;
	//AddressW = Clamp;
};

//-----------------------------------------------
// Fresnel Power
//-----------------------------------------------
half FresnelPower
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.01;
	float UIMax		= 10;
	float UIStep	= 0.1;
	string UIName	=  "Fresnel Factor";
> = 1.5;

//-----------------------------------------------
// Reflection Blur
//-----------------------------------------------
half ReflectionBlurFactor
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 10;
	float UIStep	= 0.1;
	string UIName	=  "Reflection Blur Factor";
> = 1;

//-----------------------------------------------
//Ambient Factor
//-----------------------------------------------
half AmbientFactor
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 2;
	float UIStep	= 0.1;
	string UIName	=  "Ambient Factor";
> = 1;

//-----------------------------------------------
//Ambient Blur
//-----------------------------------------------
half AmbientBlurFactor
<
	string UIType = "FloatSpinner";
	float UIMin		= 0.0;
	float UIMax		= 10;
	float UIStep	= 1;
	string UIName	=  "Ambient Blur Factor";
> = 3;

//---------------------------------------------
// structures 
//---------------------------------------------
struct APP_INPUT
{
	float4 Position : POSITION;
	float3 Normal 	: NORMAL;
	float2 TexCoord : TEXCOORD0;
	float4 Color		: TEXCOORD1;
	float3 Tangent  : TANGENT;
	float3 Binormal : BINORMAL;
};

struct vOUT2
{
	float4 Position		: POSITION;
	float4 Color			: COLOR0;
	float2 TexCoord		: TEXCOORD0;
	float3 LightDir		: TEXCOORD1;
	float3 EyeDir		: TEXCOORD2;
	float3 Tgt			: TEXCOORD3;
	float3 Bnl		 	: TEXCOORD4;
	float3 Nml		 	: TEXCOORD5;
};

//---------------------------------------------
// Vertex shader
//---------------------------------------------
vOUT2 MainVS2(APP_INPUT IN)
{
	vOUT2 OUT;

	half3 tangent  = IN.Tangent.xyz;
	half3 normal   = IN.Normal.xyz;
	half3 binormal = IN.Binormal.xyz;

	// position ,uvs
	OUT.Position	= mul(IN.Position,wvpMatrix);
	OUT.TexCoord.xy = IN.TexCoord;
	
	half3 EyePos = half3(wvMatrixI[3].xyz);

	half3 EyeDir = normalize(EyePos - IN.Position.xyz);
	OUT.EyeDir.xyz = mul(half4(EyeDir.xyz, 0.0),wMatrix).xyz;

	half3 wPos = mul(IN.Position,wMatrix).xyz;
	half3 LightDir = normalize(LightPos.xyz - wPos);
	OUT.LightDir.xyz = LightDir;

	OUT.Tgt = mul(half4(tangent, 0),wMatrix).xyz;
	OUT.Bnl = mul(half4(binormal, 0),wMatrix).xyz;
	OUT.Nml = mul(half4(normal, 0),wMatrix).xyz;

	OUT.Color = IN.Color;
	return OUT;
}

float3 texcoordEnvswizzle(float3 ref){
	return float3(ref.x,ref.z,ref.y);
}

float3 normalSwizzle(float3 ref){
	return float3(ref.y,ref.x,ref.z);
}

//---------------------------------------------
// pixel shader shader
//---------------------------------------------

float4 Test(vOUT2 IN) : COLOR
{
	return float4(1,1,1,1);
}

float4 Gloss2PS(vOUT2 IN) : COLOR
{
	half3 diffuse		= tex2D(diffuseSampler,	IN.TexCoord.xy).rgb;
	half alpha			= tex2D(AlphaSampler,	IN.TexCoord.xy).r;
	half3 normal		= tex2D(NormalSampler,	IN.TexCoord.xy).xyz * 2.0 - 1.0;
	normal = normalSwizzle(normal);
	normal = normalize(float3(normal.x * BumpFactor, normal.y * BumpFactor, normal.z));
	half specular		= SpecFactor * tex2D(SpecularSampler,IN.TexCoord.xy).r;
	half specularPower	= tex2D(SpecularPowerSampler,IN.TexCoord.xy).r;
	half emissive		= tex2D(EmissiveSampler,IN.TexCoord.xy).r;

	clip(alpha-0.25);
	
	half3 wNormal;
	wNormal.xyz  = normal.x * IN.Tgt;
	wNormal.xyz += normal.y * IN.Bnl;
	wNormal.xyz += normal.z * IN.Nml;
	wNormal = normalize(wNormal);

	// normalize
	half3 wLightDir	= normalize(IN.LightDir);
	half3 wEyeDir		= normalize(IN.EyeDir);
	
	half NdotE = dot(wNormal.xyz, wEyeDir.xyz);
	half LdotN = dot(wNormal.xyz, -wLightDir.xyz);

	// reflection
	half4 wReflect;
	wReflect.xyz = 2.0 * NdotE * wNormal.xyz - wEyeDir.xyz;	
	wReflect.w = (1.0 - GlossFactor) * specularPower.r * ReflectionBlurFactor;
	half3 env = texCUBElod(EnvSampler, half4(texcoordEnvswizzle(wReflect.xyz),wReflect.w)).rgb;

	//Fresnel effect
	half fresnel = pow(1.001-saturate(NdotE),FresnelPower);
	half specFresnel = pow(1.001-saturate(NdotE),SpecularFresnelPower);

	//BRDF
	half2 BRDF_UV = float2(saturate(NdotE+OffsetU), saturate((LdotN + 1) * 0.5) + OffsetV);
	BRDF_UV.y = 1-BRDF_UV.y;
	half3 BRDF = tex2D(BRDFSampler, BRDF_UV.xy).rgb;

	// specular
	half specPower = specularPower * (50 * GlossFactor + 1);
	half3 LdotR = saturate(pow(saturate(dot(wLightDir.xyz, wReflect.xyz)), specPower)) * LightColor.rgb;

	// ambient
	half3 ambient = AmbientFactor*texCUBElod(EnvSampler,  float4(texcoordEnvswizzle(wNormal.xyz),AmbientBlurFactor)).rgb;

	// Thin Surface translucence
	half3 BdotN = (1-alpha) * saturate(dot(wEyeDir.xyz, -wLightDir.xyz)) * saturate(LdotN) * LightColor.rgb;

	// diffuse
	half3 diffuseFull = diffuse.rgb * ( ambient.rgb + BRDF.rgb);

	return float4(IN.Color.xyz * lerp(diffuseFull.rgb, env.rgb, specular*fresnel)
	+ (specular * LdotR * specFresnel)
	+ (BdotN * diffuse.xyz * SSSColor)
	+ (EmissiveFactor * emissive * EmissiveColor) , 1.0);
}


//-------------------------------------------------
//  techniques 
//-------------------------------------------------
technique Generic
{
	pass P0
	{
		AlphaBlendEnable = FALSE;
		CullMode = CW;
		ZWriteEnable = TRUE;
		ZFunc = LESSEQUAL;
		ZEnable = TRUE;

		VertexShader   = compile vs_1_1 MainVS2();
		PixelShader = compile ps_3_0 Gloss2PS();
		
	}
}

